import sys

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.request import Requests, parse_response_object

logger = get_logger(__name__)


def get_current_remote(quiet: bool):
    client_config = load_client_config()
    remote_name = client_config.lifecycle_url
    lifecycle_url = resolve_lifecycle_url(client_config, remote_name)
    if quiet or not sys.stdout.isatty():
        print(lifecycle_url)
    elif remote_name == lifecycle_url:
        logger.info(f'Current remote is "{lifecycle_url}"')
    else:
        logger.info(f'Current remote is "{remote_name}", resolving to {lifecycle_url}')


def get_current_pub_address(quiet: bool):
    client_config = load_client_config()
    remote_name = client_config.lifecycle_url
    lifecycle_url = resolve_lifecycle_url(client_config, remote_name)

    r = Requests.get(
        f'{lifecycle_url}/api/v1/info',
    )
    response = parse_response_object(r, 'Lifecycle response error')
    pub_url = response.get('external_pub_url')
    assert pub_url, 'Pub URL is not specified by the Racetrack server'
    if quiet or not sys.stdout.isatty():
        print(pub_url)
    else:
        logger.info(f'Current Pub URL is: {pub_url}')



from pathlib import Path
import json
from typing import Dict

from jsonschema import validate

from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.load import get_manifest_path
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest.merge import load_merged_manifest
from racetrack_client.utils.datamodel import datamodel_to_yaml_str
from racetrack_client.utils.semver import SemanticVersion

logger = get_logger(__name__)


def load_validated_manifest(
    path: str,
    extra_vars: Dict[str, str] = None,
) -> Manifest:
    """
    Load and validate manifest from a path. Raise exception in case of a defect.
    :param path: path to a Job manifest file or to a directory with it
    :param extra_vars: key-value pairs overriding manifest values
    :return loaded, valid Manifest
    """
    manifest_path = get_manifest_path(path)
    manifest = load_merged_manifest(manifest_path, extra_vars or {})

    with wrap_context('validating manifest'):
        validate_manifest(manifest)

    logger.debug(f'Manifest file "{manifest_path}" is valid')
    return manifest


def validate_manifest(manifest: Manifest):
    """Check whether manifest is valid. Raise exception in case of error"""
    serialized_manifest = manifest.model_dump(exclude_none=True)
    path = Path(__file__).with_name('schema.json')
    with path.open('r') as f:
        schema = json.load(f)
        validate(serialized_manifest, schema=schema)

    with wrap_context('parsing Job version'):
        SemanticVersion(manifest.version)

    manifest.warn_if_using_deprecated_fields(logger)


def validate_and_show_manifest(
    path: str,
    extra_vars: Dict[str, str] = None,
):
    manifest = load_validated_manifest(path, extra_vars)
    logger.info(f'Manifest file "{path}" is valid')
    print(datamodel_to_yaml_str(manifest))



def format_bytes(count: int):
    """Format bytes amount as a human friendly string with KiB, MiB, GiB"""
    if count == 1:
        return '1 byte'
    elif count < 1024:
        return f'{count} bytes'
    elif count < 1024**2:
        return f'{count/1024:.2f} KiB'
    elif count < 1024**3:
        return f'{count/1024/1024:.2f} MiB'
    else:
        return f'{count/1024/1024/1024:.2f} GiB'



from pathlib import Path
import fnmatch
from typing import List, Iterable

DEFAULT_IGNORE_PATTERNS = [
    '*.zip',
    '.git',
    '*.pyc',
    '__pycache__',
    '.gitignore',
    '/.racetrackignore',
    '/plugin-manifest.yaml',
]


class FilenameMatcher:
    """
    A matcher checking if the file path matches one of the given wildcard patterns.
    This is especially useful for checking if the file should be ignored due to gitignore-alike patterns.
    Filename patterns should be given in Unix shell-style wildcards syntax
    (very similar to well-known .gitignore syntax).

    Besides, if you want to include or exclude specific files,
    you can prepend the patterns with "+" (to include) or "-" (to exclude).
    Rules are evaluated in order of occurrence. For instance:
        -whole_dir
        +whole_dir/but_this
        -whole_dir/but_this/without_that
    """

    def __init__(self, file_patterns: List[str] = None, apply_defaults: bool = True) -> None:
        self.patterns: List[str] = []

        if file_patterns:
            file_patterns = list(filter(None, file_patterns))  # filter non-empty
            self.patterns.extend(file_patterns)

        if apply_defaults:
            self.patterns.extend(DEFAULT_IGNORE_PATTERNS)

    def match_path(self, relative_path: Path) -> bool:
        """Check whether the file path should be included (True) or excluded (False)"""
        result = True

        for pattern in self.patterns:

            # result that should be applied if a file matches the pattern
            pattern_result = False
            if pattern.startswith('-'):
                pattern = pattern[1:]
            elif pattern.startswith('+'):
                pattern_result = True
                pattern = pattern[1:]

            if match_file_pattern(relative_path, pattern):
                result = pattern_result

        return result

    def list_files(self, root: Path) -> Iterable[Path]:
        """
        List all files in the given directory that match the patterns.
        Return relative paths.
        """
        yield from self._list_files_in_subdir(root, root)

    def _list_files_in_subdir(self, root: Path, directory: Path) -> Iterable[Path]:
        for path in directory.iterdir():
            relative_path = path.relative_to(root)

            if not self.match_path(relative_path):
                continue

            if path.is_dir():
                yield from self._list_files_in_subdir(root, path)
            else:
                yield relative_path


def match_file_pattern(relative_path: Path, pattern: str) -> bool:
    if pattern.startswith('/'):
        return _match_file_pattern_beginning(relative_path, pattern[1:])

    pattern_parts = Path(pattern).parts
    path_parts = relative_path.parts

    if len(pattern_parts) == 1:
        for path_part in path_parts:
            if fnmatch.fnmatchcase(path_part, pattern):
                return True

    if len(pattern_parts) > 1:
        if len(path_parts) < len(pattern_parts):
            return False
        for path_part, pattern_part in zip(reversed(path_parts), reversed(pattern_parts)):
            if not fnmatch.fnmatchcase(path_part, pattern_part):
                return False
        return True

    return False


def _match_file_pattern_beginning(relative_path: Path, pattern: str) -> bool:
    path_parts = relative_path.parts
    pattern_parts = Path(pattern).parts

    if len(path_parts) < len(pattern_parts):
        return False

    for path_part, pattern_part in zip(path_parts, pattern_parts):
        if not fnmatch.fnmatchcase(path_part, pattern_part):
            return False

    return True



import io
from pathlib import Path
from typing import Optional
import zipfile

from racetrack_client.log.logs import get_logger
from racetrack_client.log.context_error import wrap_context
from racetrack_client.plugin.bundler.filename_matcher import FilenameMatcher
from racetrack_client.plugin.plugin_manifest import PluginManifest
from racetrack_client.utils.datamodel import parse_yaml_datamodel, datamodel_to_yaml_str
from racetrack_client.utils.time import now

logger = get_logger(__name__)

PLUGIN_FILENAME = 'plugin.py'
PLUGIN_MANIFEST_FILENAME = 'plugin-manifest.yaml'


def bundle_plugin(workdir: str, out_dir: Optional[str], out_filename: Optional[str], plugin_version: Optional[str]):
    """Turn local plugin code into ZIP file"""
    plugin_dir = Path(workdir)
    assert (plugin_dir / PLUGIN_FILENAME).is_file(), f'{plugin_dir / PLUGIN_FILENAME} file was not found in a plugin directory'

    plugin_manifest_file = plugin_dir / PLUGIN_MANIFEST_FILENAME
    assert plugin_manifest_file.is_file(), f'{plugin_manifest_file} file was not found in a plugin directory'
    plugin_manifest = _load_plugin_manifest(plugin_manifest_file)

    if plugin_version:
        plugin_manifest.version = plugin_version

    out_dir_path = Path(out_dir) if out_dir else plugin_dir
    assert out_dir_path.is_dir(), f'out directory {out_dir_path} doesn\'t exist'
    if not out_filename:
        out_filename = f'{plugin_manifest.name}-{plugin_manifest.version}.zip'
    out_path = out_dir_path / out_filename

    ignore_file = plugin_dir / '.racetrackignore'
    if ignore_file.is_file():
        logger.info(f'ignoring file patterns found in {ignore_file}')
        inclusion_matcher = FilenameMatcher(ignore_file.read_text().splitlines())
    else:
        inclusion_matcher = FilenameMatcher()

    with zipfile.ZipFile(out_path.as_posix(), mode="w", compression=zipfile.ZIP_STORED) as _zip:
        for file in plugin_dir.rglob('*'):

            if file.is_dir():
                continue
            relative_path = file.relative_to(plugin_dir)
            if not inclusion_matcher.match_path(relative_path):
                continue

            logger.debug(f'writing file to zip: {relative_path}')
            _zip.write(file.as_posix(), arcname=relative_path.as_posix())

        _write_plugin_manifest(_zip, plugin_manifest)

    logger.info(f'plugin {plugin_manifest.name} has been exported to: {out_path.resolve().absolute()}')


def _load_plugin_manifest(manifest_file: Path) -> PluginManifest:
    with wrap_context(f'parsing plugin manfiest from {manifest_file}'):
        yaml_str = manifest_file.read_text()
        return parse_yaml_datamodel(yaml_str, PluginManifest)


def _write_plugin_manifest(_zip: zipfile.ZipFile, plugin_manifest: PluginManifest):
    logger.debug(f'writing plugin manifest to zip: {PLUGIN_MANIFEST_FILENAME}')
    plugin_manifest.build_date = now().strftime('%Y-%m-%dT%H%M%SZ')
    manifest_output = io.StringIO()
    manifest_output.write(datamodel_to_yaml_str(plugin_manifest))
    _zip.writestr(PLUGIN_MANIFEST_FILENAME, manifest_output.getvalue())



import json as json_module
import ssl
from email.message import Message
from http.client import HTTPResponse
from http.client import responses
from typing import Any, List, Optional, Dict, Union
from urllib import request
from urllib.error import HTTPError, URLError
import urllib.parse as urlparse
from urllib.parse import urlencode

from racetrack_client.log.context_error import ContextError

DEBUG_MODE = False  # print the actual bytes sent in requests and responses


class RequestError(ContextError):
    """HTTP Request error due to network error, bad URL, unreachable address"""
    def __init__(self, error_context: str):
        super().__init__(error_context)


class ResponseError(RequestError):
    """HTTP error due to bad response code"""
    def __init__(self, error_context: str, status_code: int):
        super().__init__(error_context)
        self.status_code = status_code


class Response:

    def __init__(
        self,
        url: str,
        method: str,
        status_code: int,
        content: bytes,
        headers: Message,
    ):
        self._url: str = url
        self._method: str = method.upper()
        self._status_code: int = status_code
        self._content: bytes = content
        self._headers: Message = headers

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def ok(self) -> bool:
        return self._status_code < 400

    def raise_for_status(self):
        if 400 <= self._status_code < 600:
            raise ResponseError(f'{self._status_code} {self.status_reason} for url: {self.url}', self._status_code)

    @property
    def status_reason(self) -> str:
        return responses[self._status_code]

    @property
    def content(self) -> bytes:
        return self._content

    @property
    def text(self) -> str:
        return self._content.decode('utf8')

    def json(self) -> Union[Dict, List]:
        response_data = self._content.decode('utf8')
        if not response_data:
            raise RuntimeError('no content in response to decode as JSON')
        return json_module.loads(response_data)

    @property
    def is_json(self) -> bool:
        return 'application/json' in self._headers['content-type']

    @property
    def url(self) -> str:
        return self._url
    
    @property
    def method(self) -> str:
        return self._method
    
    @property
    def headers(self) -> Message:
        return self._headers
    
    def header(self, name: str) -> Optional[str]:
        return self._headers[name]

    @property
    def error_details(self) -> str:
        return f'{self.status_code} {self.status_reason} for url: {self.method} {self.url}'


class Requests:

    insecure: bool = True  # whether to verify SSL certificates

    @classmethod
    def get(
            cls,
            url: str,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        return cls._make_request('GET', url, None, None, params, headers, timeout)

    @classmethod
    def post(
            cls,
            url: str,
            json: Optional[Any] = None,
            data: Optional[bytes] = None,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        return cls._make_request('POST', url, json, data, params, headers, timeout)

    @classmethod
    def put(
            cls,
            url: str,
            json: Optional[Any] = None,
            data: Optional[bytes] = None,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        return cls._make_request('PUT', url, json, data, params, headers, timeout)

    @classmethod
    def delete(
            cls,
            url: str,
            json: Optional[Any] = None,
            data: Optional[bytes] = None,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        return cls._make_request('DELETE', url, json, data, params, headers, timeout)

    @classmethod
    def request(
            cls,
            method: str,
            url: str,
            json: Optional[Any] = None,
            data: Optional[bytes] = None,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        return cls._make_request(method.upper(), url, json, data, params, headers, timeout)

    @classmethod
    def _make_request(
            cls,
            method: str,
            url: str,
            jsondata: Optional[Any] = None,
            data: Optional[bytes] = None,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            ) -> Response:
        """
        Make HTTP request and return response object.
        :param method: HTTP method: GET, POST, PUT, DELETE
        :param url: URL starting with http(s)://
        :param jsondata: payload object to be serialized as JSON content
        :param data: raw payload bytes to be sent as HTTP content
        :param params: query params (at the end of path)
        :param headers: request headers
        :param timeout: timeout in seconds for blocking operations like the connection attempt
        (if not specified, the global default timeout setting will be used)
        :throws RequestError: if the request fails
        """
        if params:
            url = build_url_with_params(url, params)

        req = request.Request(url, method=method)
        kwargs: Dict[str, Any] = {}
        
        if headers:
            for name, value in headers.items():
                if value is not None:
                    req.add_header(name, value)

        if jsondata is not None:
            jsondataasbytes = json_module.dumps(jsondata).encode('utf-8')
            req.add_header('Content-Type', 'application/json; charset=utf-8')
            req.add_header('Content-Length', str(len(jsondataasbytes)))
            kwargs['data'] = jsondataasbytes
        elif data is not None:
            req.add_header('Content-Length', str(len(data)))
            kwargs['data'] = data
        
        if not req.has_header('Accept'):
            req.add_header('Accept', 'application/json, */*')

        if timeout is not None:
            kwargs['timeout'] = timeout

        if not req.has_header('User-Agent'):
            req.add_header('User-Agent', 'request')

        kwargs['context'] = cls._get_ssl_context()

        if DEBUG_MODE:
            http_handler = request.HTTPHandler(debuglevel=2)
            https_handler = request.HTTPSHandler(context=kwargs['context'], debuglevel=2)
            opener = request.build_opener(http_handler, https_handler)
            request.install_opener(opener)
            # Passing 'context' argument to urllib.request.urlopen rebuilds the URL opener,
            # not giving a chance to turn the debug mode on.
            del kwargs['context']

        try:
            http_response: HTTPResponse = request.urlopen(req, **kwargs)
            return Response(
                url=url,
                method=method,
                status_code=http_response.status,
                content=http_response.read(),
                headers=http_response.headers,
            )
        except HTTPError as e:
            return Response(
                url=url,
                method=method,
                status_code=e.code,
                content=e.read(),
                headers=e.headers,
            )
        except URLError as e:
            raise RequestError(f'Request {method} {url} failed') from e
        except BaseException as e:
            raise RequestError(f'Request to {url} failed') from e

    @classmethod
    def _get_ssl_context(cls) -> ssl.SSLContext:
        ctx = ssl.create_default_context()
        if cls.insecure:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        return ctx


def build_url_with_params(
    base_url: str,
    params: Dict[str, Any],
) -> str:
    url_parts = list(urlparse.urlparse(base_url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)
    

def parse_response(response: Response, error_context: str) -> Optional[Union[Dict, List]]:
    """
    Ensure response was successful. If not, try to extract error message from it.
    :return: response parsed as JSON object

    This function handles all of these following cases:
    - if request went fine, return parsed JSON object,
    - if request went fine, but response is empty (eg. 204 status after DELETing), return None
    - if request went fine, but there was eg. JSON parsing error,
     raise "Deployment error: Json is invalid in character 5"
    - if request completely failed with empty response (eg. due to Kubernetes issues),
    raise "Deployment error: 404 Not found for url http://localhost/blahblah"
    - if request failed gracefully from other RT component returning {"error": "you have no power here"},
    raise "Deployment error: 500 Internal Server Error: you have no power here"
    """
    try:
        result: Optional[Union[Dict, List]] = None
        if 'application/json' in response.headers['content-type']:
            result = response.json()

        if response.ok:
            return result

        if result is not None and isinstance(result, dict) and 'error' in result:
            raise ContextError(response.status_reason, RuntimeError(result.get("error")))
        raise ResponseError(response.error_details, response.status_code)
    except Exception as e:
        raise ResponseError(error_context, response.status_code) from e


def parse_response_object(response: Response, error_context: str) -> Dict:
    try:
        if 'application/json' not in response.headers['content-type']:
            raise RuntimeError(f'expected JSON response, got "{response.headers["content-type"]}", '
                               f'{response.error_details}')

        result = response.json()

        if response.ok:
            assert isinstance(result, dict), f'response JSON expected to be a dictionary, got {type(result)}'
            return result

        if result is not None and isinstance(result, dict) and 'error' in result:
            raise ContextError(response.status_reason, RuntimeError(result.get("error")))
        raise ResponseError(response.error_details, response.status_code)
    except Exception as e:
        raise ResponseError(error_context, response.status_code) from e


def parse_response_list(response: Response, error_context: str) -> List:
    try:
        if 'application/json' not in response.headers['content-type']:
            raise RuntimeError(f'expected JSON response, got "{response.headers["content-type"]}", '
                               f'{response.error_details}')

        result = response.json()

        if response.ok:
            assert isinstance(result, list), f'response JSON expected to be a list, got {type(result)}'
            return result

        if result is not None and isinstance(result, dict) and 'error' in result:
            raise ContextError(response.status_reason, RuntimeError(result.get("error")))
        raise ResponseError(response.error_details, response.status_code)
    except Exception as e:
        raise ResponseError(error_context, response.status_code) from e



import os
from pathlib import Path
import sys
import tempfile
from typing import List, Optional, Dict
from urllib.parse import urlsplit
import tarfile
import io
from base64 import b64encode
from enum import Enum

import backoff

from racetrack_client.client.env import read_secret_vars, SecretVars
from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.client_config import ClientConfig, Credentials
from racetrack_client.client_config.io import load_client_config
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.load import get_manifest_path
from racetrack_client.manifest.merge import load_merged_manifest_dict
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.plugin.bundler.filename_matcher import FilenameMatcher
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.byte import format_bytes
from racetrack_client.utils.request import parse_response_object, Requests
from racetrack_client.log.logs import get_logger


logger = get_logger(__name__)

DEPLOYMENT_TIMEOUT_SECS = 60 * 60  # 1 hour


class DeploymentError(RuntimeError):
    def __init__(self, cause: BaseException):
        super().__init__()
        self.__cause__ = cause

    def __str__(self):
        return f'deployment error: {self.__cause__}'


class BuildContextMethod(str, Enum):
    local = "local"
    git = "git"
    default = "default"


def catch_piped_input():
    """Check for piped input and read it if available, avoiding indefinite blocking."""
    piped_data, file_path= None, None
    if not sys.stdin.isatty() and (piped_data := sys.stdin.read()):
        current_workdir = os.getcwd()
        with tempfile.NamedTemporaryFile(mode='w+', dir=current_workdir, delete=False) as tmp_file:
            tmp_file.write(piped_data)
            tmp_file.flush()
            file_path = tmp_file.name
    return file_path


def send_deploy_request(
    workdir: str,
    client_config: Optional[ClientConfig] = None,
    lifecycle_url: Optional[str] = None,
    force: bool = False,
    build_context_method: BuildContextMethod = BuildContextMethod.default,
    extra_vars: Dict[str, str] = None,
    build_flags: List[str] = [],
):
    """
    Send request deploying a new Job to running Lifecycle instance
    :param workdir: directory with job.yaml manifest
    :param client_config: client configuration to use (if not the defaullt)
    :param lifecycle_url: Racetrack server's URL or alias name
    :param force: overwrite existing job without asking
    :param build_context_method: decides whether to build from local files or from git:
        local - build an image from local build context,
        git - build from git repository,
        None - apply default strategy:
            if working on local dev, local build context gets activated, otherwise git
    :param extra_vars: key-value pairs overriding manifest values
    """
    # Creates a tmp file to store the piped input. Have to try/finally to ensure it gets deleted
    try:
        # If piped input is available, use it as the workdir
        tmp_file = catch_piped_input()
        workdir = tmp_file or workdir

        if client_config is None:
            client_config = load_client_config()
        manifest: Manifest = load_validated_manifest(workdir, extra_vars)
        manifest_dict: Dict = load_merged_manifest_dict(get_manifest_path(workdir), extra_vars)
        logger.debug(f'Manifest loaded: {manifest}')

        lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
        user_auth = get_user_auth(client_config, lifecycle_url)

        logger.info(f'deploying workspace "{workdir}" (job {manifest.name} v{manifest.version}) to {lifecycle_url}')

        build_context = get_build_context(lifecycle_url, workdir, manifest, build_context_method)
        git_credentials = get_git_credentials(manifest, client_config)
        secret_vars = read_secret_vars(workdir, manifest)

        # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::DeployEndpoint` for server-side implementation
        r = Requests.post(
            f'{lifecycle_url}/api/v1/deploy',
            json=get_deploy_request_payload(manifest_dict, git_credentials, secret_vars, build_context, force, build_flags),
            headers=get_auth_request_headers(user_auth),
        )
        response = parse_response_object(r, 'Lifecycle deploying error')
        deploy_id = response["id"]
        logger.info(f'job deployment requested: {deploy_id}')

        try:
            job_url = _wait_for_deployment_result(lifecycle_url, deploy_id, user_auth, [], [])
        except Exception as e:
            raise DeploymentError(e)
        logger.info(f'Job "{manifest.name}" has been deployed. Check out {job_url} to access your Job')
    finally:
        if tmp_file is not None and os.path.exists(tmp_file) and os.path.isfile(tmp_file):
            os.remove(tmp_file)


def get_deploy_request_payload(
    manifest_dict: Dict,
    git_credentials: Optional[Credentials],
    secret_vars: SecretVars,
    build_context: Optional[str],
    force: bool,
    build_flags: List[str],
) -> Dict:
    return {
        "manifest": manifest_dict,
        "git_credentials": git_credentials.model_dump() if git_credentials is not None else None,
        "secret_vars": secret_vars.model_dump(),
        "build_context": build_context,
        "force": force,
        "build_flags": build_flags,
    }


@backoff.on_exception(
    backoff.fibo, TimeoutError, max_value=3, max_time=DEPLOYMENT_TIMEOUT_SECS, jitter=None, logger=None
)
def _wait_for_deployment_result(lifecycle_url: str, deploy_id: str, user_auth: str, phases: List[Optional[str]], warningss: List[Optional[str]]) -> str:
    # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::DeployIdEndpoint` for server-side implementation
    r = Requests.get(
        f'{lifecycle_url}/api/v1/deploy/{deploy_id}',
        headers=get_auth_request_headers(user_auth),
    )
    response = parse_response_object(r, 'Lifecycle deployment status')
    warnings = response.get('warnings')
    if not warningss or warningss[-1] != warnings:  # don't print the same warnings again
        warningss.append(warnings)
        if warnings:
            logger.warning(warnings)

    status = response['status'].lower()
    if status == 'failed':
        raise RuntimeError(response['error'])
    if status == 'done':
        return response.get('job', {}).get('pub_url', '')

    phase = response.get('phase')
    if not phases or phases[-1] != phase:  # don't print the same phase again
        phases.append(phase)
        if phase:
            logger.info(f'deployment in progress: {phase}...')
        else:
            logger.info(f'deployment in progress...')

    raise TimeoutError('Deployment timeout error')


def get_git_credentials(manifest: Manifest, client_config: ClientConfig) -> Optional[Credentials]:
    if manifest.git.remote in client_config.git_credentials:
        return client_config.git_credentials[manifest.git.remote]

    split = urlsplit(manifest.git.remote)

    url = f'{split.scheme}://{split.netloc}{split.path}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.scheme}://{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    return None


def get_build_context(lifecycle_url: str, workdir: str, manifest: Manifest, build_context_method: BuildContextMethod) -> Optional[str]:
    """
    Use compressed local build context in order to prevent the need to commit every change.
    """
    if build_context_method == BuildContextMethod.default:
        build_context_method = _determine_default_build_context_method(lifecycle_url, manifest)
    if build_context_method == BuildContextMethod.local:
        return encode_build_context(workdir)
    else:
        return None


def _determine_default_build_context_method(lifecycle_url: str, manifest: Manifest) -> BuildContextMethod:
    """
    Decide whether to use local build context or to fetch source from git.
    It avoids fetching from remote git when using samples from local project repository.
    """
    if _is_url_localhost(lifecycle_url) and manifest.owner_email == 'sample@example.com':
        logger.info('using local build context due to deploying to localhost')
        return BuildContextMethod.local
    return BuildContextMethod.git


def _is_url_localhost(url: str) -> bool:
    split = urlsplit(url)
    return split.hostname in {'localhost', '127.0.0.1', '::1'}


def encode_build_context(workdir: str) -> str:
    """Compress build context directory to .tar.gz and encode it to base64 string"""
    tar_fileobj = io.BytesIO()
    workdir_path = Path(workdir)
    if workdir_path.is_file():
        workdir_path = workdir_path.parent

    ignore_file = workdir_path / '.gitignore'
    if ignore_file.is_file():
        logger.debug(f'skipping file patterns found in {ignore_file}')
        inclusion_matcher = FilenameMatcher(ignore_file.read_text().splitlines())
    else:
        inclusion_matcher = FilenameMatcher()

    files_count = 0
    with tarfile.open(fileobj=tar_fileobj, mode='w:gz') as tar:
        for relative_path in inclusion_matcher.list_files(workdir_path):
            absolute: Path = workdir_path / relative_path
            tar.add(str(absolute), arcname=str(relative_path))
            files_count += 1
    encoded = b64encode(tar_fileobj.getvalue()).decode()
    logger.info(f"local build context encoded: {files_count} files, {format_bytes(len(encoded))}")
    return encoded



import io
import subprocess
import sys
import threading
from pathlib import Path
from typing import Optional, Callable

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


def shell_output(
    cmd: str, 
    workdir: Optional[Path] = None, 
    print_stdout: bool = False,
    print_log: bool = True,
    read_bytes: bool = False,
    output_filename: Optional[str] = None,
) -> str:
    """
    Run system shell command and return its output.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    :param cmd: shell command to run
    :param workdir: working directory for the command
    :param print_stdout: whether to print stdout from a subprocess to the main process stdout
    :param print_log: whether to print a log message about running the command
    :param read_bytes: whether to read raw bytes from the subprocess stdout instead of whole lines
    :param output_filename: file to write the output in real time
    """
    captured_stream = shell(cmd, workdir,
                            print_stdout=print_stdout, print_log=print_log, output_filename=output_filename,
                            read_bytes=read_bytes, raw_output=False)
    return captured_stream.getvalue()


def shell(
    cmd: str, 
    workdir: Optional[Path] = None, 
    print_stdout: bool = True,
    print_log: bool = True,
    raw_output: bool = False,
    output_filename: Optional[str] = None,
    read_bytes: bool = False,
) -> io.StringIO:
    """
    Run system shell command.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    :param cmd: shell command to run
    :param workdir: working directory for the command
    :param print_stdout: whether to print stdout from a subprocess to the main process stdout
    :param print_log: whether to print a log message about running the command
    :param raw_output: whether to let subprocess manage stdout/stderr on its own instead of capturing it
    :param output_filename: file to write the output in real time
    :param read_bytes: whether to read raw bytes from the subprocess stdout instead of whole lines
    :raises:
        CommandError: in case of non-zero command exit code.
    """
    if print_log:
        logger.debug(f'Command: {cmd}')
    if len(cmd) > 4096:  # see https://github.com/torvalds/linux/blob/v5.11/drivers/tty/n_tty.c#L1681
        raise RuntimeError('maximum tty line length has been exceeded')
    output_file = open(output_filename, 'a') if output_filename else None

    if raw_output:
        process = subprocess.Popen(cmd, stdout=None, stderr=None, shell=True, cwd=workdir)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=workdir)

    try:
        # fork command output to stdout, captured buffer and output file
        captured_stream = io.StringIO()
        if raw_output:
            process.wait()
            if process.returncode != 0:
                raise CommandError(cmd, '', process.returncode)
            return captured_stream

        if read_bytes:
            while True:
                chunk: bytes = process.stdout.read(1)
                if chunk == b'':
                    break
                chunk_str = chunk.decode()

                if print_stdout:
                    sys.stdout.write(chunk_str)
                    sys.stdout.flush()
                if output_file is not None:
                    output_file.write(chunk_str)
                captured_stream.write(chunk_str)

        else:
            for line in iter(process.stdout.readline, b''):
                line_str = line.decode()

                if print_stdout:
                    sys.stdout.write(line_str)
                    sys.stdout.flush()
                if output_file is not None:
                    output_file.write(line_str)
                captured_stream.write(line_str)

        process.wait()
        if output_file is not None:
            output_file.close()
        if process.returncode != 0:
            stdout = captured_stream.getvalue()
            raise CommandError(cmd, stdout, process.returncode)
        return captured_stream
    except KeyboardInterrupt:
        logger.warning('killing subprocess')
        process.kill()
        raise


class CommandOutputStream:
    def __init__(self,
                 cmd: str,
                 on_next_line: Callable[[str], None],
                 on_error: Callable[['CommandError'], None] = None,
                 workdir: Optional[Path] = None,
                 print_stdout: bool = False):
        """
        Manager of a shell command process running in background,
        monitoring stdout lines coming from it and streaming them.
        :param cmd: shell command
        :param on_next_line: callback function called whenever the new stdout line appears
        :param on_error: callback function called when process ends with non-zero return code
        :param workdir: working directory to call command with
        :param print_stdout: whether to write output also to stdout
        """
        self.stop = False
        logger.debug(f'Command: {cmd}')
        self.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=workdir)

        def monitor_output(stream):
            captured_stream = io.StringIO()
            for line in iter(stream.process.stdout.readline, b''):
                if stream.stop:
                    break
                line_str = line.decode()
                if print_stdout:
                    sys.stdout.write(line_str)
                    sys.stdout.flush()
                on_next_line(line_str)
                captured_stream.write(line_str)

            stream.process.wait()
            if stream.process.returncode != 0 and on_error is not None:
                stdout = captured_stream.getvalue()
                on_error(CommandError(cmd, stdout, stream.process.returncode))

        self.thread = threading.Thread(
            target=monitor_output,
            args=(self,),
            daemon=True,
        )
        self.thread.start()

    def interrupt(self):
        self.stop = True
        self.process.terminate()


class CommandError(RuntimeError):
    def __init__(self, cmd: str, stdout: str, returncode: int):
        super().__init__()
        self.cmd = cmd
        self.stdout = stdout
        self.returncode = returncode

    def __str__(self):
        return f'command failed: {self.cmd}: {self.stdout}'



import os
import traceback
from typing import Collection, Iterable, Tuple

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)

PRINT_FULL_TRACEBACK = False


def log_exception(e: BaseException):
    """
    Log exception in concise one-line format containing message, exception type and short traceback
    """
    try:
        error_message, cause, tb = exception_details(e)
        logger.error(error_message, extra={'cause': cause, 'traceback': tb})
    except BaseException as e:
        logger.exception(str(e))


def exception_details(e: BaseException) -> Tuple[str, str, str]:
    """
    Return concise details of exception: message, root cause type and one-line details traceback
    :param e: Exception instance
    :return Tuple of: message string, root cause type and traceback
    """
    traceback_ex = traceback.TracebackException(type(e), e, e.__traceback__, limit=None)
    traceback_lines = list(_get_traceback_lines(traceback_ex))
    traceback_str = ', '.join(traceback_lines)
    cause = _root_cause_type(e)
    error_msg = str(e).strip()
    if PRINT_FULL_TRACEBACK:
        traceback_str += '\n' + '\n'.join(traceback.format_exception(e))
    return error_msg, cause, traceback_str


def short_exception_details(e: BaseException) -> str:
    """
    Return concise one-line details of exception, containing message, traceback and root cause type
    """
    error_message, cause, tb = exception_details(e)
    return f'{error_message}, cause={cause}, traceback={tb}'


def _root_cause_type(e: BaseException) -> str:
    while e.__cause__ is not None:
        e = e.__cause__
    return type(e).__name__


def _get_traceback_lines(t1: traceback.TracebackException) -> Iterable[str]:
    while True:
        frames: Collection[traceback.FrameSummary] = t1.stack
        for frame in frames:
            yield f'{os.path.normpath(frame.filename)}:{frame.lineno}'

        if t1.__cause__ is None:
            break
        t1 = t1.__cause__



from typing import Optional, List, Dict, Any

from pydantic import BaseModel, ConfigDict, Field

from racetrack_client.utils.quantity import AnnotatedQuantity

from logging import Logger

class GitManifest(BaseModel):
    model_config = ConfigDict(extra='forbid')

    # URL of git remote: HTTPS, SSH or directory path to a remote repository
    remote: str
    branch: Optional[str] = None
    # subdirectory relative to git repo root
    directory: str = '.'


class ResourcesManifest(BaseModel):
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True)

    # minimum memory amount in bytes, eg. 256Mi
    memory_min: Optional[AnnotatedQuantity] = None
    # maximum memory amount in bytes, eg. 1Gi
    memory_max: Optional[AnnotatedQuantity] = None
    # minimum CPU consumption in cores, eg. 10m
    cpu_min: Optional[AnnotatedQuantity] = None
    # maximum CPU consumption in cores, eg. 1000m
    cpu_max: Optional[AnnotatedQuantity] = None


class Manifest(BaseModel):
    """Job Manifest file - build recipe to get deployable image from source code workspace"""
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True, populate_by_name=True)

    # name of the Job Workload
    name: str

    # email address of the Job's owner to reach out
    owner_email: str

    git: GitManifest

    # version of the Job
    version: str = '0.0.1'

    # Jobtype wrapper used to embed model
    jobtype: Optional[str] = None
    lang: Optional[str] = None  # Deprecated

    # relative path to base manifest file, which will be extended by this manifest
    extends: Optional[str] = None

    # type of deployed image: docker image, packer, AMI
    image_type: str = 'docker'

    # system-wide packages that should be installed with apt
    system_dependencies: Optional[List[str]] = None

    # env vars for building
    build_env: Optional[Dict[str, str]] = None
    # env vars for runtime
    runtime_env: Optional[Dict[str, str]] = None
    # secret env vars loaded from an external file applied on building
    secret_build_env_file: Optional[str] = None
    # secret env vars loaded from an external file applied at runtime
    secret_runtime_env_file: Optional[str] = None

    # labels - job metadata for humans
    labels: Optional[Dict[str, Any]] = None

    # list of public job endpoints that can be accessed without authentication
    public_endpoints: Optional[List[str]] = None

    # number of instances of the Job
    replicas: int = 1

    # resources demands to allocate to the Job
    resources: Optional[ResourcesManifest] = None

    # Extra parameters specified by the jobtype
    jobtype_extra: Optional[Dict[str, Any]] = None
    golang: Optional[Dict[str, Any]] = None  # Deprecated
    python: Optional[Dict[str, Any]] = None  # Deprecated
    docker: Optional[Dict[str, Any]] = None  # Deprecated
    wrapper_properties: Optional[Dict[str, Any]] = None  # Deprecated

    # Back-end platform where to deploy the service
    infrastructure_target: Optional[str] = None

    # original YAML string from which the manifest was parsed, field for internal use only
    origin_yaml_: Optional[str] = Field(None, exclude=True)
    # original dictionary from which the manifest was parsed, field for internal use only
    origin_dict_: Optional[Dict[str, Any]] = Field(None, exclude=True)

    # This is the "source of truth for deprications", schema.json and DeployForm.vue has to follow this.
    def get_deprecated_fields(self):
        return {
            'lang': 'jobtype:',
            'golang': 'jobtype_extra:',
            'python': 'jobtype_extra:',
            'docker': 'jobtype_extra:',
            'wrapper_properties': 'jobtype_extra:'
        }

    def get_jobtype(self) -> str:
        return self.jobtype if self.jobtype else self.lang

    def get_jobtype_extra(self) -> Dict[str, Any]:
        for field in [self.jobtype_extra, self.golang, self.python, self.wrapper_properties]:
            if field is not None:
                return field
        return {}

    def warn_if_using_deprecated_fields(self, logger: Logger):
        for field, replacement in self.get_deprecated_fields().items():
            if getattr(self, field) is not None:
                logger.warning(f'{field}: is deprecated. Use {replacement}: instead.')



import contextlib
import threading
from typing import Callable, Dict

import socketio

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


class LogsConsumer:
    def __init__(self,
                 url: str,
                 socketio_path: str,
                 resource_properties: Dict[str, str],
                 on_next_line: Callable[[str], None]):
        """
        Consumer of log lines stream.
        :param url: base URL of Socket.IO server
        :param socketio_path: The endpoint where the Socket.IO server is installed.
        :param resource_properties: metadata of a resource being monitored (properties to identify a Job)
        :param on_next_line: callback function called whenever a new line is consumed.
        """
        self.url = url
        self.socketio_path = socketio_path
        self.resource_properties = resource_properties
        self.sio = socketio.Client(reconnection=True, ssl_verify=False)

        @self.sio.event
        def connect():
            logger.debug('Socket.IO connection established')

        @self.sio.event
        def connect_error(data):
            logger.error(f'Socket.IO connection error: {data}')

        @self.sio.event
        def disconnect():
            logger.debug('Disconnected from Socket.IO server')

        @self.sio.event
        def logs_nextline(data):
            """"Event for retrieveing a new log line"""
            on_next_line(data.get('line'))
            return True

    def _subscribe_for_logs(self):
        self.sio.call('subscribe_for_logs', {
            'resource_properties': self.resource_properties,
        })

    @contextlib.contextmanager
    def connect_async(self):
        """Set up and tear down a client context consuming logs in background"""
        self.sio.connect(self.url, socketio_path=self.socketio_path)
        self._subscribe_for_logs()
        try:
            threading.Thread(
                target=lambda: self.sio.wait(),
                daemon=True,
            ).start()
            yield
        finally:
            self.sio.disconnect()

    def connect_and_wait(self):
        """Connect to a server synchronously and listen for the logs"""
        try:
            self.sio.connect(self.url, socketio_path=self.socketio_path)
            self._subscribe_for_logs()
            self.sio.wait()
        except KeyboardInterrupt:
            self.sio.disconnect()
